from ... import platforms
from .. import state
from . import cpu

# ARM32 has a number of variants with very odd overlaps. Rather than risk
# copy-paste errors as I figure out who gets which control registers, I've
# implemented each subsystem variant in its own mixin.


class ARM(cpu.CPU):
    """Base class for ARM 32-bit CPU models.

    All ARM CPUs share the same basic registers, but there are at least two
    dimensions of difference for the available modes.
    """

    def __init__(self):
        super().__init__()
        # *** General-purpose registers ***
        self.r0 = state.Register("r0", size=4)
        self.add(self.r0)
        self.r1 = state.Register("r1", size=4)
        self.add(self.r1)
        self.r2 = state.Register("r2", size=4)
        self.add(self.r2)
        self.r3 = state.Register("r3", size=4)
        self.add(self.r3)
        self.r4 = state.Register("r4", size=4)
        self.add(self.r4)
        self.r5 = state.Register("r5", size=4)
        self.add(self.r5)
        self.r6 = state.Register("r6", size=4)
        self.add(self.r6)
        self.r7 = state.Register("r7", size=4)
        self.add(self.r7)
        self.r8 = state.Register("r8", size=4)
        self.add(self.r8)
        # r9 doubles as the Static base pointer
        self.r9 = state.Register("r9", size=4)
        self.add(self.r9)
        self.sb = state.RegisterAlias("sb", self.r9, size=4, offset=0)
        self.add(self.sb)
        # r10 doubles as the Stack Limit pointer
        self.r10 = state.Register("r10", size=4)
        self.add(self.r10)
        self.sl = state.RegisterAlias("sl", self.r10, size=4, offset=0)
        self.add(self.sl)
        # r11 doubles as the Frame Pointer, if desired.
        self.r11 = state.Register("r11", size=4)
        self.add(self.r11)
        self.fp = state.RegisterAlias("fp", self.r11, size=4, offset=0)
        self.add(self.fp)
        # r12 doubles as the Intra-call scratch register
        self.r12 = state.Register("r12", size=4)
        self.add(self.r12)
        self.ip = state.RegisterAlias("ip", self.r12, size=4, offset=0)
        self.add(self.ip)
        self.sp = state.Register("sp", size=4)
        self.add(self.sp)
        self.lr = state.Register("lr", size=4)
        self.add(self.lr)
        self.pc = state.Register("pc", size=4)
        self.add(self.pc)


class ARMCPUMixinM:
    """Abstract class for M-series CPUs.

    The main difference between M and R/A
    is the available system status registers.
    """

    def __init__(self):
        super().__init__()
        # *** Special Registers ***
        # Program Status Register
        # NOTE: PSR can be accessed through several masked aliases.
        # These are read-only, so I'm not including them.
        # - apsr: Just the condition flags
        # - ipsr: Just exception information
        # - epsr: Just execution state info
        # - iapsr: apsr | ipsr
        # - eapsr: apsr | epsr
        # - iepsr: ipsr | epsr
        # - xpsr: apsr | ipsr | epsr
        #
        # NOTE: Unicorn doesn't have a model for PSR, only its aliases
        self.psr = state.Register("psr", size=4)
        self.add(self.psr)
        # Exception Mask Register
        self.primask = state.Register("primask", size=4)
        self.add(self.primask)
        # Base Priority Mask Register
        self.basepri = state.Register("basepri", size=4)
        self.add(self.basepri)
        # Fault Mask Register
        self.faultmask = state.Register("faultmask", size=4)
        self.add(self.faultmask)
        # Control register; includes a lot of flags.
        self.control = state.Register("control", size=4)
        self.add(self.control)

        # *** Stack Pointer Bank ***
        # sp is actually an alias to one of these two.
        # Exactly which one depends on a bit in control.
        # Emulators that care should be careful when loading state.

        # Main Stack Pointer
        self.msp = state.Register("msp", size=4)
        self.add(self.msp)
        # Process Stack Pointer
        self.psp = state.Register("psp", size=4)
        self.add(self.psp)


class ARMCPUMixinRA:
    """Mixin for R- or A-series CPUs.

    The main difference between M and R/A
    is the available system status registers.
    """

    def __init__(self):
        super().__init__()
        # *** Special Registers ***
        # Current Program Status Register
        # NOTE: CPSR can be accessed through several masked aliases.
        # These are read-only, so I'm not including them.
        # - isetstate: Just includes instruction set control bits
        # - itstate: Just includes state bits for Thumb IT instruction
        self.cpsr = state.Register("cpsr", size=4)
        self.add(self.cpsr)
        # Saved Program Status Register
        self.spsr = state.Register("spsr", size=4)
        self.add(self.spsr)

        # *** Register Banks ***
        # sp, lr, and spsr are actually aliases to one of these.
        # Which one they reference depends on execution mode.
        # Emulators that care should be careful when loading state.
        # NOTE: Use User-mode copies of registers unless the mode has its own.

        # User-mode Stack Pointer
        self.sp_usr = state.Register("sp_usr", size=4)
        self.add(self.sp_usr)
        # User-mode Link Register
        self.lr_usr = state.Register("lr_usr", size=4)
        self.add(self.lr_usr)
        # User-mode r8
        self.r8_usr = state.Register("r8_usr", size=4)
        self.add(self.r8_usr)
        # User-mode r9
        self.r9_usr = state.Register("r9_usr", size=4)
        self.add(self.r9_usr)
        # User-mode r10
        self.r10_usr = state.Register("r10_usr", size=4)
        self.add(self.r10_usr)
        # User-mode r11
        self.r11_usr = state.Register("r11_usr", size=4)
        self.add(self.r11_usr)
        # User-mode r12
        self.r12_usr = state.Register("r12_usr", size=4)
        self.add(self.r12_usr)

        # Hypervisor Stack Pointer
        self.sp_hyp = state.Register("sp_hyp", size=4)
        self.add(self.sp_hyp)
        # Hypervisor Saved PSR
        self.spsr_hyp = state.Register("spsr_hyp", size=4)
        self.add(self.spsr_hyp)
        # Hypervisor Exception Link Register
        # NOTE: This isn't so much banked, as it only exists in hypervisor mode.
        self.elr_hyp = state.Register("elr_hyp", size=4)
        self.add(self.elr_hyp)

        # Supervisor Stack Pointer
        self.sp_svc = state.Register("sp_svc", size=4)
        self.add(self.sp_svc)
        # Supervisor Link Register
        self.lr_svc = state.Register("lr_svc", size=4)
        self.add(self.lr_svc)
        # Supervisor Saved PSR
        self.spsr_svc = state.Register("spsr_svc", size=4)
        self.add(self.spsr_svc)

        # Abort-state Stack Pointer
        self.sp_abt = state.Register("sp_abt", size=4)
        self.add(self.sp_abt)
        # Abort-state Link Register
        self.lr_abt = state.Register("lr_abt", size=4)
        self.add(self.lr_abt)
        # Abort-state Saved PSR
        self.spsr_abt = state.Register("spsr_abt", size=4)
        self.add(self.spsr_abt)

        # Undefined-mode Stack Pointer
        self.sp_und = state.Register("sp_und", size=4)
        self.add(self.sp_und)
        # Undefined-mode Link Register
        self.lr_und = state.Register("lr_und", size=4)
        self.add(self.lr_und)
        # Undefined-mode Saved PSR
        self.spsr_und = state.Register("spsr_und", size=4)
        self.add(self.spsr_und)

        # Monitor-mode Stack Pointer
        self.sp_mon = state.Register("sp_mon", size=4)
        self.add(self.sp_mon)
        # Monitor-mode Link Register
        self.lr_mon = state.Register("lr_mon", size=4)
        self.add(self.lr_mon)
        # Monitor-mode Saved PSR
        self.spsr_mon = state.Register("spsr_mon", size=4)
        self.add(self.spsr_mon)

        # IRQ-mode Stack Pointer
        self.sp_irq = state.Register("sp_irq", size=4)
        self.add(self.sp_irq)
        # IRQ-mode Link Register
        self.lr_irq = state.Register("lr_irq", size=4)
        self.add(self.lr_irq)
        # IRQ-mode Saved PSR
        self.spsr_irq = state.Register("spsr_irq", size=4)
        self.add(self.spsr_irq)

        # FIQ-mode Stack Pointer
        self.sp_fiq = state.Register("sp_fiq", size=4)
        self.add(self.sp_fiq)
        # FIQ-mode Link Register
        self.lr_fiq = state.Register("lr_fiq", size=4)
        self.add(self.lr_fiq)
        # FIQ-mode Saved PSR
        self.spsr_fiq = state.Register("spsr_fiq", size=4)
        self.add(self.spsr_fiq)
        # FIQ-mode r8
        self.r8_fiq = state.Register("r8_fiq", size=4)
        self.add(self.r8_fiq)
        # FIQ-mode r9
        self.r9_fiq = state.Register("r9_fiq", size=4)
        self.add(self.r9_fiq)
        # FIQ-mode r10
        self.r10_fiq = state.Register("r10_fiq", size=4)
        self.add(self.r10_fiq)
        # FIQ-mode r11
        self.r11_fiq = state.Register("r11_fiq", size=4)
        self.add(self.r11_fiq)
        # FIQ-mode r12
        self.r12_fiq = state.Register("r12_fiq", size=4)
        self.add(self.r12_fiq)


class ARMCPUMixinFPEL:
    """Mixin for little-endian ARM CPUs with FP extensions

    This is one kind of floating-point extension
    which offers 64-bit scalar operations
    """

    def __init__(self):
        super().__init__()
        # *** Floating point control registers ***
        # Floating-point Status and Control Register
        self.fpscr = state.Register("fpscr", size=4)
        self.add(self.fpscr)
        # Floating-point Exception Control Register
        self.fpexc = state.Register("fpexc", size=4)
        self.add(self.fpexc)
        # Floating-point System ID Register
        self.fpsid = state.Register("fpsid", size=4)
        self.add(self.fpsid)
        # Media and VFP Feature Register 0
        self.mvfr0 = state.Register("mvfr0", size=4)
        self.add(self.mvfr0)
        # Media and VFP Feature Register 1
        self.mvfr1 = state.Register("mvfr1", size=4)
        self.add(self.mvfr1)

        # *** Floating point registers ***
        self.d0 = state.Register("d0", size=8)
        self.add(self.d0)
        self.s0 = state.RegisterAlias("s0", self.d0, size=4, offset=0)
        self.add(self.s0)
        self.s1 = state.RegisterAlias("s1", self.d0, size=4, offset=4)
        self.add(self.s1)
        self.d1 = state.Register("d1", size=8)
        self.add(self.d1)
        self.s2 = state.RegisterAlias("s2", self.d1, size=4, offset=0)
        self.add(self.s2)
        self.s3 = state.RegisterAlias("s3", self.d1, size=4, offset=4)
        self.add(self.s3)
        self.d2 = state.Register("d2", size=8)
        self.add(self.d2)
        self.s4 = state.RegisterAlias("s4", self.d2, size=4, offset=0)
        self.add(self.s4)
        self.s5 = state.RegisterAlias("s5", self.d2, size=4, offset=4)
        self.add(self.s5)
        self.d3 = state.Register("d3", size=8)
        self.add(self.d3)
        self.s6 = state.RegisterAlias("s6", self.d3, size=4, offset=0)
        self.add(self.s6)
        self.s7 = state.RegisterAlias("s7", self.d3, size=4, offset=4)
        self.add(self.s7)
        self.d4 = state.Register("d4", size=8)
        self.add(self.d4)
        self.s8 = state.RegisterAlias("s8", self.d4, size=4, offset=0)
        self.add(self.s8)
        self.s9 = state.RegisterAlias("s9", self.d4, size=4, offset=4)
        self.add(self.s9)
        self.d5 = state.Register("d5", size=8)
        self.add(self.d5)
        self.s10 = state.RegisterAlias("s10", self.d5, size=4, offset=0)
        self.add(self.s10)
        self.s11 = state.RegisterAlias("s11", self.d5, size=4, offset=4)
        self.add(self.s11)
        self.d6 = state.Register("d6", size=8)
        self.add(self.d6)
        self.s12 = state.RegisterAlias("s12", self.d6, size=4, offset=0)
        self.add(self.s12)
        self.s13 = state.RegisterAlias("s13", self.d6, size=4, offset=4)
        self.add(self.s13)
        self.d7 = state.Register("d7", size=8)
        self.add(self.d7)
        self.s14 = state.RegisterAlias("s14", self.d7, size=4, offset=0)
        self.add(self.s14)
        self.s15 = state.RegisterAlias("s15", self.d7, size=4, offset=4)
        self.add(self.s15)
        self.d8 = state.Register("d8", size=8)
        self.add(self.d8)
        self.s16 = state.RegisterAlias("s16", self.d8, size=4, offset=0)
        self.add(self.s16)
        self.s17 = state.RegisterAlias("s17", self.d8, size=4, offset=4)
        self.add(self.s17)
        self.d9 = state.Register("d9", size=8)
        self.add(self.d9)
        self.s18 = state.RegisterAlias("s18", self.d9, size=4, offset=0)
        self.add(self.s18)
        self.s19 = state.RegisterAlias("s19", self.d9, size=4, offset=4)
        self.add(self.s19)
        self.d10 = state.Register("d10", size=8)
        self.add(self.d10)
        self.s20 = state.RegisterAlias("s20", self.d10, size=4, offset=0)
        self.add(self.s20)
        self.s21 = state.RegisterAlias("s21", self.d10, size=4, offset=4)
        self.add(self.s21)
        self.d11 = state.Register("d11", size=8)
        self.add(self.d11)
        self.s22 = state.RegisterAlias("s22", self.d11, size=4, offset=0)
        self.add(self.s22)
        self.s23 = state.RegisterAlias("s23", self.d11, size=4, offset=4)
        self.add(self.s23)
        self.d12 = state.Register("d12", size=8)
        self.add(self.d12)
        self.s24 = state.RegisterAlias("s24", self.d12, size=4, offset=0)
        self.add(self.s24)
        self.s25 = state.RegisterAlias("s25", self.d12, size=4, offset=4)
        self.add(self.s25)
        self.d13 = state.Register("d13", size=8)
        self.add(self.d13)
        self.s26 = state.RegisterAlias("s26", self.d13, size=4, offset=0)
        self.add(self.s26)
        self.s27 = state.RegisterAlias("s27", self.d13, size=4, offset=4)
        self.add(self.s27)
        self.d14 = state.Register("d14", size=8)
        self.add(self.d14)
        self.s28 = state.RegisterAlias("s28", self.d14, size=4, offset=0)
        self.add(self.s28)
        self.s29 = state.RegisterAlias("s29", self.d14, size=4, offset=4)
        self.add(self.s29)
        self.d15 = state.Register("d15", size=8)
        self.add(self.d15)
        self.s30 = state.RegisterAlias("s30", self.d15, size=4, offset=0)
        self.add(self.s30)
        self.s31 = state.RegisterAlias("s31", self.d15, size=4, offset=4)
        self.add(self.s31)


class ARMCPUMixinVFPEL:
    """Mixin for little-endian ARM CPUs with VFP/NEON mixins

    This is one kind of floating-point extension
    which supports up to 128-bit scalar and SIMD vector operations.

    VFP and NEON are always optional extensions;
    The two can exist independently, and VFP can support either
    16 or 32 double registers.
    This is the maximal set of registers, assuming both are supported.
    """

    def __init__(self):
        super().__init__()
        # *** Floating-point Control Registers ***
        # Floating-point Status and Control Register
        self.fpscr = state.Register("fpscr", size=4)
        self.add(self.fpscr)
        # Floating-point Exception Control Register
        self.fpexc = state.Register("fpexc", size=4)
        self.add(self.fpexc)
        # Floating-point System ID Register
        self.fpsid = state.Register("fpsid", size=4)
        self.add(self.fpsid)
        # Media and VFP Feature Register 0
        self.mvfr0 = state.Register("mvfr0", size=4)
        self.add(self.mvfr0)
        # Media and VFP Feature Register 1
        self.mvfr1 = state.Register("mvfr1", size=4)
        self.add(self.mvfr1)
        # *** Floating-point Registers ****
        self.q0 = state.Register("q0", size=16)
        self.add(self.q0)
        self.d0 = state.RegisterAlias("d0", self.q0, size=8, offset=0)
        self.add(self.d0)
        self.s0 = state.RegisterAlias("s0", self.q0, size=4, offset=0)
        self.add(self.s0)
        self.s1 = state.RegisterAlias("s1", self.q0, size=4, offset=4)
        self.add(self.s1)
        self.d1 = state.RegisterAlias("d1", self.q0, size=8, offset=8)
        self.add(self.d1)
        self.s2 = state.RegisterAlias("s2", self.q0, size=4, offset=8)
        self.add(self.s2)
        self.s3 = state.RegisterAlias("s3", self.q0, size=4, offset=12)
        self.add(self.s3)
        self.q1 = state.Register("q1", size=16)
        self.add(self.q1)
        self.d2 = state.RegisterAlias("d2", self.q1, size=8, offset=0)
        self.add(self.d2)
        self.s4 = state.RegisterAlias("s4", self.q1, size=4, offset=0)
        self.add(self.s4)
        self.s5 = state.RegisterAlias("s5", self.q1, size=4, offset=4)
        self.add(self.s5)
        self.d3 = state.RegisterAlias("d3", self.q1, size=8, offset=8)
        self.add(self.d3)
        self.s6 = state.RegisterAlias("s6", self.q1, size=4, offset=8)
        self.add(self.s6)
        self.s7 = state.RegisterAlias("s7", self.q1, size=4, offset=12)
        self.add(self.s7)
        self.q2 = state.Register("q2", size=16)
        self.add(self.q2)
        self.d4 = state.RegisterAlias("d4", self.q2, size=8, offset=0)
        self.add(self.d4)
        self.s8 = state.RegisterAlias("s8", self.q2, size=4, offset=0)
        self.add(self.s8)
        self.s9 = state.RegisterAlias("s9", self.q2, size=4, offset=4)
        self.add(self.s9)
        self.d5 = state.RegisterAlias("d5", self.q2, size=8, offset=8)
        self.add(self.d5)
        self.s10 = state.RegisterAlias("s10", self.q2, size=4, offset=8)
        self.add(self.s10)
        self.s11 = state.RegisterAlias("s11", self.q2, size=4, offset=12)
        self.add(self.s11)
        self.q3 = state.Register("q3", size=16)
        self.add(self.q3)
        self.d6 = state.RegisterAlias("d6", self.q3, size=8, offset=0)
        self.add(self.d6)
        self.s12 = state.RegisterAlias("s12", self.q3, size=4, offset=0)
        self.add(self.s12)
        self.s13 = state.RegisterAlias("s13", self.q3, size=4, offset=4)
        self.add(self.s13)
        self.d7 = state.RegisterAlias("d7", self.q3, size=8, offset=8)
        self.add(self.d7)
        self.s14 = state.RegisterAlias("s14", self.q3, size=4, offset=8)
        self.add(self.s14)
        self.s15 = state.RegisterAlias("s15", self.q3, size=4, offset=12)
        self.add(self.s15)
        self.q4 = state.Register("q4", size=16)
        self.add(self.q4)
        self.d8 = state.RegisterAlias("d8", self.q4, size=8, offset=0)
        self.add(self.d8)
        self.s16 = state.RegisterAlias("s16", self.q4, size=4, offset=0)
        self.add(self.s16)
        self.s17 = state.RegisterAlias("s17", self.q4, size=4, offset=4)
        self.add(self.s17)
        self.d9 = state.RegisterAlias("d9", self.q4, size=8, offset=8)
        self.add(self.d9)
        self.s18 = state.RegisterAlias("s18", self.q4, size=4, offset=8)
        self.add(self.s18)
        self.s19 = state.RegisterAlias("s19", self.q4, size=4, offset=12)
        self.add(self.s19)
        self.q5 = state.Register("q5", size=16)
        self.add(self.q5)
        self.d10 = state.RegisterAlias("d10", self.q5, size=8, offset=0)
        self.add(self.d10)
        self.s20 = state.RegisterAlias("s20", self.q5, size=4, offset=0)
        self.add(self.s20)
        self.s21 = state.RegisterAlias("s21", self.q5, size=4, offset=4)
        self.add(self.s21)
        self.d11 = state.RegisterAlias("d11", self.q5, size=8, offset=8)
        self.add(self.d11)
        self.s22 = state.RegisterAlias("s22", self.q5, size=4, offset=8)
        self.add(self.s22)
        self.s23 = state.RegisterAlias("s23", self.q5, size=4, offset=12)
        self.add(self.s23)
        self.q6 = state.Register("q6", size=16)
        self.add(self.q6)
        self.d12 = state.RegisterAlias("d12", self.q6, size=8, offset=0)
        self.add(self.d12)
        self.s24 = state.RegisterAlias("s24", self.q6, size=4, offset=0)
        self.add(self.s24)
        self.s25 = state.RegisterAlias("s25", self.q6, size=4, offset=4)
        self.add(self.s25)
        self.d13 = state.RegisterAlias("d13", self.q6, size=8, offset=8)
        self.add(self.d13)
        self.s26 = state.RegisterAlias("s26", self.q6, size=4, offset=8)
        self.add(self.s26)
        self.s27 = state.RegisterAlias("s27", self.q6, size=4, offset=12)
        self.add(self.s27)
        self.q7 = state.Register("q7", size=16)
        self.add(self.q7)
        self.d14 = state.RegisterAlias("d14", self.q7, size=8, offset=0)
        self.add(self.d14)
        self.s28 = state.RegisterAlias("s28", self.q7, size=4, offset=0)
        self.add(self.s28)
        self.s29 = state.RegisterAlias("s29", self.q7, size=4, offset=4)
        self.add(self.s29)
        self.d15 = state.RegisterAlias("d15", self.q7, size=8, offset=8)
        self.add(self.d15)
        self.s30 = state.RegisterAlias("s30", self.q7, size=4, offset=8)
        self.add(self.s30)
        self.s31 = state.RegisterAlias("s31", self.q7, size=4, offset=12)
        self.add(self.s31)
        # NOTE: This isn't a typo; there are only 32 single-precision sX registers
        # This does mean that only half the VFP register space can be used
        # for single-precision arithmetic.
        self.q8 = state.Register("q8", size=16)
        self.add(self.q8)
        self.d16 = state.RegisterAlias("d16", self.q8, size=8, offset=0)
        self.add(self.d16)
        self.d17 = state.RegisterAlias("d17", self.q8, size=8, offset=8)
        self.add(self.d17)
        self.q9 = state.Register("q9", size=16)
        self.add(self.q9)
        self.d18 = state.RegisterAlias("d18", self.q9, size=8, offset=0)
        self.add(self.d18)
        self.d19 = state.RegisterAlias("d19", self.q9, size=8, offset=8)
        self.add(self.d19)
        self.q10 = state.Register("q10", size=16)
        self.add(self.q10)
        self.d20 = state.RegisterAlias("d20", self.q10, size=8, offset=0)
        self.add(self.d20)
        self.d21 = state.RegisterAlias("d21", self.q10, size=8, offset=8)
        self.add(self.d21)
        self.q11 = state.Register("q11", size=16)
        self.add(self.q11)
        self.d22 = state.RegisterAlias("d22", self.q11, size=8, offset=0)
        self.add(self.d22)
        self.d23 = state.RegisterAlias("d23", self.q11, size=8, offset=8)
        self.add(self.d23)
        self.q12 = state.Register("q12", size=16)
        self.add(self.q12)
        self.d24 = state.RegisterAlias("d24", self.q12, size=8, offset=0)
        self.add(self.d24)
        self.d25 = state.RegisterAlias("d25", self.q12, size=8, offset=8)
        self.add(self.d25)
        self.q13 = state.Register("q13", size=16)
        self.add(self.q13)
        self.d26 = state.RegisterAlias("d26", self.q13, size=8, offset=0)
        self.add(self.d26)
        self.d27 = state.RegisterAlias("d27", self.q13, size=8, offset=8)
        self.add(self.d27)
        self.q14 = state.Register("q14", size=16)
        self.add(self.q14)
        self.d28 = state.RegisterAlias("d28", self.q14, size=8, offset=0)
        self.add(self.d28)
        self.d29 = state.RegisterAlias("d29", self.q14, size=8, offset=8)
        self.add(self.d29)
        self.q15 = state.Register("q15", size=16)
        self.add(self.q15)
        self.d30 = state.RegisterAlias("d30", self.q15, size=8, offset=0)
        self.add(self.d30)
        self.d31 = state.RegisterAlias("d31", self.q15, size=8, offset=8)
        self.add(self.d31)


class ARMv5T(ARMCPUMixinM, ARM):
    """CPU Model for ARMv5t little-endian."""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V5T, platforms.Byteorder.LITTLE
    )


class ARMv6M(ARMCPUMixinFPEL, ARMCPUMixinM, ARM):
    """CPU Model for ARMv6-M little-endian."""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V6M, platforms.Byteorder.LITTLE
    )


class ARMv6MThumb(ARMv6M):
    """CPU Model for ARMv6-M little-endian in thumb mode."""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V6M_THUMB, platforms.Byteorder.LITTLE
    )


class ARMv7M(ARMCPUMixinFPEL, ARMCPUMixinM, ARM):
    """CPU Model for ARMv7-M little-endian."""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V7M, platforms.Byteorder.LITTLE
    )


class ARMv7R(ARMCPUMixinVFPEL, ARMCPUMixinRA, ARM):
    """CPU Model for ARMv7-R little-endian."""

    # TODO: v7r and v7a have different MMUs, which I don't implement yet.
    platform = platforms.Platform(
        platforms.Architecture.ARM_V7R, platforms.Byteorder.LITTLE
    )


class ARMv7A(ARMCPUMixinVFPEL, ARMCPUMixinRA, ARM):
    """CPU Model for ARMv7-A little-endian."""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V7A, platforms.Byteorder.LITTLE
    )



import ctypes

# TODO: Figure out how to type-hint a ctypes objects
# The base class for ctypes objects is private (with good reason.)
# There are hacky and likely-fragile ways to get it.


class TypedPointer(ctypes.c_void_p):
    """Typed pointer class

    This class fills a gap in ctypes ability to represent data.

    The existing typed pointer classes are designed
    as direct references to Python objects;
    it's not possible to set a specific address.

    c_void_p represents the correct kind of value,
    but has no associated data type.

    Warning:
        Do not instantiate this class directly! The referenced type needs to be
        bound to a class, since ctypes uses instances to represent specific
        variables or fields. Use ``create_typed_pointer()`` to create a
        subclass for your type.
    """

    _type = None

    def __init__(self, *args, **kwargs):
        # No idea what the signature for this should be,
        # and don't want to pick one in case it changes
        if self.__class__ == TypedPointer:
            raise TypeError("Cannot instantiate TypedPointer directly")
        # NOTE: Due to a bug, can't use super() in ctypes subclasses
        ctypes.c_void_p.__init__(self, *args, **kwargs)

    @property
    def type(self):
        """The type referenced by this pointer."""

        return self._type

    def __str__(self):
        return f"Pointer {self.type} = {self.value}"

    def __eq__(self, other):
        return (
            isinstance(other, TypedPointer)
            and other.type == self.type
            and other.value == self.value
        )


_pointertypes = {
    None: ctypes.c_void_p,
    ctypes.c_char: ctypes.c_char_p,
    ctypes.c_wchar: ctypes.c_wchar_p,
}


def create_typed_pointer(reference):
    """Create a typed pointer class.

    The referenced type should be any ctypes type definition, or ``None`` to
    represent 'void'.

    Referenced types that already have a ctypes pointer value type will return
    that type, not a ``TypedPointer``::

        create_typed_pointer(c_char)  # returns c_char_p
        create_typed_pointer(c_wchar)  # returns c_wchar_p
        create_typed_pointer(None)  # returns c_void_p

    Arguments:
        reference: The ctypes object defining the referenced type.

    Returns:
        A subclass of ``TypedPointer`` representing your referenced type.
    """
    if reference in _pointertypes:
        return _pointertypes[reference]
    else:
        # Dynamically create a new subclass of TypedPointer to represent this
        # particular type
        name = f"{type.__name__}Pointer"
        cls = type(name, (TypedPointer,), {"type": reference})
        _pointertypes[reference] = cls
        return cls


__all__ = [
    "TypedPointer",
    "create_typed_pointer",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import MIPSELSysVModel


class MIPSELSysVBasename(Basename, MIPSELSysVModel):
    pass


class MIPSELSysVDirname(Dirname, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVBasename",
    "MIPSELSysVDirname",
]



from . import (
    aarch64,
    amd64,
    armel,
    armhf,
    i386,
    loongarch64,
    mips,
    mips64,
    mips64el,
    mipsel,
    powerpc,
    riscv64,
)
from .mmio import MemoryMappedModel
from .model import *  # noqa: F401, F403
from .model import __all__ as __model__

__all__ = __model__ + [
    "MemoryMappedModel",
    "aarch64",
    "amd64",
    "armel",
    "armhf",
    "i386",
    "loongarch64",
    "mips",
    "mipsel",
    "mips64",
    "mips64el",
    "powerpc",
    "riscv64",
]



from ....platforms import Architecture, Byteorder
from .machdef import PandaMachineDef


class i386MachineDef(PandaMachineDef):
    arch = Architecture.X86_32
    byteorder = Byteorder.LITTLE

    panda_arch = "i386"

    # I'm going to define all the ones we are making possible as of now
    # I need to submit a PR to change to X86 32 bit and to includ eflags
    _registers_general = {"eax", "ebx", "ecx", "edx", "esi", "edi", "esp", "ebp", "eip"}
    _registers_short = {"ax", "bx", "cx", "dx", "si", "di", "sp", "bp"}
    _registers_byte = {"al", "bl", "cl", "dl", "ah", "bh", "ch", "dh"}
    _registers_flags = {"eflags", "flags"}
    _registers_seg = {"es", "cs", "ss", "ds", "fs", "gs"}
    _registers_control = {"cr0", "cr1", "cr2", "cr3", "cr4"}
    _registers_debug = {"dr0", "dr1", "dr2", "dr3", "dr6", "dr7"}
    _registers_mmr = {"gdtr": "gdt", "idtr": "idt", "tr": "tr", "ldtr": "ldt"}
    _registers_x87 = {
        "fpr0",
        "fpr1",
        "fpr2",
        "fpr3",
        "fpr4",
        "fpr5",
        "fpr6",
        "fpr7",
        "fctrl",
        "fstat",
        "ftag",
        "fip",
        "fdp",
        "fop",
    }
    _registers_mmx = {"mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7"}
    _registers_xmm = {"xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7"}
    _registers_pc = {"pc": "eip", "ip": None}
    _registers_absent = {"dil", "sil", "spl", "bpl", "cr8"}

    _registers = {}
    _registers = _registers | {i: i for i in _registers_general}
    _registers = _registers | {i: i for i in _registers_short}
    _registers = _registers | {i: i for i in _registers_byte}
    _registers = _registers | {i: None for i in _registers_flags}
    _registers = _registers | {i: i for i in _registers_seg}
    _registers = _registers | {i: i for i in _registers_control}
    _registers = _registers | {i: None for i in _registers_debug}
    _registers = _registers | {i: j for i, j in _registers_mmr.items()}
    _registers = _registers | {i: None for i in _registers_x87}
    _registers = _registers | {i: None for i in _registers_mmx}
    _registers = _registers | {i: None for i in _registers_xmm}
    _registers = _registers | {i: j for i, j in _registers_pc.items()}
    _registers = _registers | {i: None for i in _registers_absent}
    # _registers = (
    #    _registers_general | _registers_byte | _registers_seg | _registers_control
    # )
    # _registers = {**_registers_general, **_register_short, **_registers_seg, **_registers_control}



from ... import platforms
from .. import state
from ..x86_registers import X86MMRRegister
from . import cpu


class I386(cpu.CPU):
    """i386 CPU state model."""

    platform = platforms.Platform(
        platforms.Architecture.X86_32, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__()
        # *** General Purpose Registers ***
        self.eax = state.Register("eax", 4)
        self.add(self.eax)
        self.ax = state.RegisterAlias("ax", self.eax, 2, 0)
        self.add(self.ax)
        self.al = state.RegisterAlias("al", self.eax, 1, 0)
        self.add(self.al)
        self.ah = state.RegisterAlias("ah", self.eax, 1, 1)
        self.add(self.ah)

        self.ebx = state.Register("ebx", 4)
        self.add(self.ebx)
        self.bx = state.RegisterAlias("bx", self.ebx, 2, 0)
        self.add(self.bx)
        self.bl = state.RegisterAlias("bl", self.ebx, 1, 0)
        self.add(self.bl)
        self.bh = state.RegisterAlias("bh", self.ebx, 1, 1)
        self.add(self.bh)

        self.ecx = state.Register("ecx", 4)
        self.add(self.ecx)
        self.cx = state.RegisterAlias("cx", self.ecx, 2, 0)
        self.add(self.cx)
        self.cl = state.RegisterAlias("cl", self.ecx, 1, 0)
        self.add(self.cl)
        self.ch = state.RegisterAlias("ch", self.ecx, 1, 1)
        self.add(self.ch)

        self.edx = state.Register("edx", 4)
        self.add(self.edx)
        self.dx = state.RegisterAlias("dx", self.edx, 2, 0)
        self.add(self.dx)
        self.dl = state.RegisterAlias("dl", self.edx, 1, 0)
        self.add(self.dl)
        self.dh = state.RegisterAlias("dh", self.edx, 1, 1)
        self.add(self.dh)

        self.esi = state.Register("esi", 4)
        self.add(self.esi)
        self.si = state.RegisterAlias("si", self.esi, 2, 0)
        self.add(self.si)
        self.sil = state.RegisterAlias("sil", self.esi, 1, 0)
        self.add(self.sil)

        self.edi = state.Register("edi", 4)
        self.add(self.edi)
        self.di = state.RegisterAlias("di", self.edi, 2, 0)
        self.add(self.di)
        self.dil = state.RegisterAlias("dil", self.edi, 1, 0)
        self.add(self.dil)

        self.ebp = state.Register("ebp", 4)
        self.add(self.ebp)
        self.bp = state.RegisterAlias("bp", self.ebp, 2, 0)
        self.add(self.bp)
        self.bpl = state.RegisterAlias("bpl", self.ebp, 1, 0)
        self.add(self.bpl)

        self.esp = state.Register("esp", 4)
        self.add(self.esp)
        self.sp = state.RegisterAlias("sp", self.esp, 2, 0)
        self.add(self.sp)
        self.spl = state.RegisterAlias("spl", self.esp, 1, 0)
        self.add(self.spl)

        # *** Instruction Pointer ***
        self.eip = state.Register("eip", 4)
        self.add(self.eip)
        self.ip = state.RegisterAlias("ip", self.eip, 2, 0)
        self.add(self.ip)

        self.pc = state.RegisterAlias("pc", self.eip, 4, 0)
        self.add(self.pc)

        # *** Segment Registers ***
        self.cs = state.Register("cs", 2)
        self.add(self.cs)
        self.ss = state.Register("ss", 2)
        self.add(self.ss)
        self.ds = state.Register("ds", 2)
        self.add(self.ds)
        self.es = state.Register("es", 2)
        self.add(self.es)
        self.fs = state.Register("fs", 2)
        self.add(self.fs)
        self.gs = state.Register("gs", 2)
        self.add(self.gs)

        # *** Flags Registers ***
        self.eflags = state.Register("eflags", 4)
        self.add(self.eflags)
        self.flags = state.RegisterAlias("flags", self.eflags, 2)
        self.add(self.flags)

        # *** Control Registers ***
        self.cr0 = state.Register("cr0", 4)
        self.add(self.cr0)
        self.cr1 = state.Register("cr1", 4)
        self.add(self.cr1)
        self.cr2 = state.Register("cr2", 4)
        self.add(self.cr2)
        self.cr3 = state.Register("cr3", 4)
        self.add(self.cr3)
        self.cr4 = state.Register("cr4", 4)
        self.add(self.cr4)
        # NOTE: I've got conflicting reports whether cr8 exists in i386.
        self.cr8 = state.Register("cr8", 4)
        self.add(self.cr8)

        # *** Debug Registers ***
        self.dr0 = state.Register("dr0", 4)
        self.add(self.dr0)
        self.dr1 = state.Register("dr1", 4)
        self.add(self.dr1)
        self.dr2 = state.Register("dr2", 4)
        self.add(self.dr2)
        self.dr3 = state.Register("dr3", 4)
        self.add(self.dr3)
        self.dr6 = state.Register("dr6", 4)
        self.add(self.dr6)
        self.dr7 = state.Register("dr7", 4)
        self.add(self.dr7)

        # *** Descriptor Table Registers
        # NOTE: Yes, this is 6 bytes; 2 byte segment selector plus 4 byte offset
        self.gdtr = X86MMRRegister("gdtr", 6)
        self.add(self.gdtr)
        self.idtr = X86MMRRegister("idtr", 6)
        self.add(self.idtr)
        self.ldtr = X86MMRRegister("ldtr", 6)
        self.add(self.ldtr)

        # *** Task Register ***
        # NOTE: Yes, this is 6 bytes; 2 byte segment selector plus 4 byte offset
        self.tr = X86MMRRegister("tr", 6)
        self.add(self.tr)

        # *** x87 registers ***
        self.fpr0 = state.Register("fpr0", 10)
        self.add(self.fpr0)
        self.fpr1 = state.Register("fpr1", 10)
        self.add(self.fpr1)
        self.fpr2 = state.Register("fpr2", 10)
        self.add(self.fpr2)
        self.fpr3 = state.Register("fpr3", 10)
        self.add(self.fpr3)
        self.fpr4 = state.Register("fpr4", 10)
        self.add(self.fpr4)
        self.fpr5 = state.Register("fpr5", 10)
        self.add(self.fpr5)
        self.fpr6 = state.Register("fpr6", 10)
        self.add(self.fpr6)
        self.fpr7 = state.Register("fpr7", 10)
        self.add(self.fpr7)

        # x87 Control Register
        self.fctrl = state.Register("fctrl", 2)
        self.add(self.fctrl)
        # x87 Status Register
        self.fstat = state.Register("fstat", 2)
        self.add(self.fstat)
        # x87 Tag Register
        self.ftag = state.Register("ftag", 2)
        self.add(self.ftag)
        # x87 Last Instruction Register
        self.fip = state.Register("fip", 8)
        self.add(self.fip)
        # x87 Last Operand Pointer
        self.fdp = state.Register("fdp", 8)
        self.add(self.fdp)
        # x87 Last Opcode
        self.fop = state.Register("fop", 2)
        self.add(self.fop)

        # NOTE: Docs disagree on the format of fip and fdp.
        # One source describes them as 48-bit offset-plus-segment,
        # the other describes them as 64-bit.
        # There may also be separate segment registers.
        # If you care about the x87 debug info, please feel free to update.

        # *** MMX Registers ***
        # NOTE: The MMX registers are aliases for the low 8 bytes of the x87 registers.
        # The two subsystems cannot be used simultaneously.
        self.mm0 = state.RegisterAlias("mm0", self.fpr0, 8, 0)
        self.add(self.mm0)
        self.mm1 = state.RegisterAlias("mm1", self.fpr1, 8, 0)
        self.add(self.mm1)
        self.mm2 = state.RegisterAlias("mm2", self.fpr2, 8, 0)
        self.add(self.mm2)
        self.mm3 = state.RegisterAlias("mm3", self.fpr3, 8, 0)
        self.add(self.mm3)
        self.mm4 = state.RegisterAlias("mm4", self.fpr4, 8, 0)
        self.add(self.mm4)
        self.mm5 = state.RegisterAlias("mm5", self.fpr5, 8, 0)
        self.add(self.mm5)
        self.mm6 = state.RegisterAlias("mm6", self.fpr6, 8, 0)
        self.add(self.mm6)
        self.mm7 = state.RegisterAlias("mm7", self.fpr7, 8, 0)
        self.add(self.mm7)

        # *** SSE Registers ***
        self.xmm0 = state.Register("xmm0", 16)
        self.add(self.xmm0)
        self.xmm1 = state.Register("xmm1", 16)
        self.add(self.xmm1)
        self.xmm2 = state.Register("xmm2", 16)
        self.add(self.xmm2)
        self.xmm3 = state.Register("xmm3", 16)
        self.add(self.xmm3)
        self.xmm4 = state.Register("xmm4", 16)
        self.add(self.xmm4)
        self.xmm5 = state.Register("xmm5", 16)
        self.add(self.xmm5)
        self.xmm6 = state.Register("xmm6", 16)
        self.add(self.xmm6)
        self.xmm7 = state.Register("xmm7", 16)
        self.add(self.xmm7)



from ....c99.signal import Raise, Signal
from ..systemv import ArmELSysVModel


class ArmELSysVRaise(Raise, ArmELSysVModel):
    pass


class ArmELSysVSignal(Signal, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVRaise",
    "ArmELSysVSignal",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import RiscV64SysVModel


class RiscV64SysVBasename(Basename, RiscV64SysVModel):
    pass


class RiscV64SysVDirname(Dirname, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVBasename",
    "RiscV64SysVDirname",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import I386SysVModel


class I386SysVBasename(Basename, I386SysVModel):
    pass


class I386SysVDirname(Dirname, I386SysVModel):
    pass


__all__ = [
    "I386SysVBasename",
    "I386SysVDirname",
]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import ArmHFSysVModel


class ArmHFSysVMemcpy(Memcpy, ArmHFSysVModel):
    pass


class ArmHFSysVMemmove(Memmove, ArmHFSysVModel):
    pass


class ArmHFSysVStrcpy(Strcpy, ArmHFSysVModel):
    pass


class ArmHFSysVStrncpy(Strncpy, ArmHFSysVModel):
    pass


class ArmHFSysVStrcat(Strcat, ArmHFSysVModel):
    pass


class ArmHFSysVStrncat(Strncat, ArmHFSysVModel):
    pass


class ArmHFSysVMemcmp(Memcmp, ArmHFSysVModel):
    pass


class ArmHFSysVStrncmp(Strncmp, ArmHFSysVModel):
    pass


class ArmHFSysVStrcmp(Strcmp, ArmHFSysVModel):
    pass


class ArmHFSysVStrcoll(Strcoll, ArmHFSysVModel):
    pass


class ArmHFSysVStrxfrm(Strxfrm, ArmHFSysVModel):
    pass


class ArmHFSysVMemchr(Memchr, ArmHFSysVModel):
    pass


class ArmHFSysVStrchr(Strchr, ArmHFSysVModel):
    pass


class ArmHFSysVStrcspn(Strcspn, ArmHFSysVModel):
    pass


class ArmHFSysVStrpbrk(Strpbrk, ArmHFSysVModel):
    pass


class ArmHFSysVStrrchr(Strrchr, ArmHFSysVModel):
    pass


class ArmHFSysVStrspn(Strspn, ArmHFSysVModel):
    pass


class ArmHFSysVStrstr(Strstr, ArmHFSysVModel):
    pass


class ArmHFSysVStrtok(Strtok, ArmHFSysVModel):
    pass


class ArmHFSysVMemset(Memset, ArmHFSysVModel):
    pass


class ArmHFSysVStrerror(Strerror, ArmHFSysVModel):
    pass


class ArmHFSysVStrlen(Strlen, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVMemcpy",
    "ArmHFSysVMemmove",
    "ArmHFSysVStrcpy",
    "ArmHFSysVStrncpy",
    "ArmHFSysVStrcat",
    "ArmHFSysVStrncat",
    "ArmHFSysVMemcmp",
    "ArmHFSysVStrncmp",
    "ArmHFSysVStrcmp",
    "ArmHFSysVStrcoll",
    "ArmHFSysVStrxfrm",
    "ArmHFSysVMemchr",
    "ArmHFSysVStrchr",
    "ArmHFSysVStrcspn",
    "ArmHFSysVStrpbrk",
    "ArmHFSysVStrrchr",
    "ArmHFSysVStrspn",
    "ArmHFSysVStrstr",
    "ArmHFSysVStrtok",
    "ArmHFSysVMemset",
    "ArmHFSysVStrerror",
    "ArmHFSysVStrlen",
]



import abc
import typing

from ghidra.app.plugin.processors.sleigh import SleighLanguageProvider
from ghidra.program.model.lang import Language, LanguageID, Register

from .... import exceptions, platforms, utils


class GhidraMachineDef:
    """Container class for ghdra architecture-specific definitions"""

    @property
    @abc.abstractmethod
    def arch(self) -> platforms.Architecture:
        """The architecture ID"""
        raise NotImplementedError("This is an abstract method")

    @property
    @abc.abstractmethod
    def byteorder(self) -> platforms.Byteorder:
        """The byte order"""
        raise NotImplementedError("This is an abstract method")

    @property
    @abc.abstractmethod
    def language_id(self) -> str:
        """The Pcode language ID"""
        raise NotImplementedError("This is an abstract method")

    # Does Pcode support single-instruction stepping for this ISA.
    #
    # Instructions with delay slots can't be lifted into Pcode
    # without also lifting the instruction in the delay slot.
    #
    # This flag indicates that this machine uses such instructions,
    # and is not safe to step in this manner
    supports_single_step: bool = True

    _registers: typing.Dict[str, typing.Optional[str]]

    def __init__(self):
        # Load the Pcode language definition
        slp = SleighLanguageProvider.getSleighLanguageProvider()
        langID = LanguageID(self.language_id)
        self.language: Language = slp.getLanguage(langID)
        assert self.language is not None

    def pcode_reg(self, name: str) -> Register:
        if name not in self._registers:
            raise KeyError(f"Unknown register for {self.arch}:{self.byteorder}: {name}")
        if self._registers[name] is None:
            raise exceptions.UnsupportedRegisterError(
                f"Register {name} not recognized by pcode for {self.arch}:{self.byteorder}"
            )
        reg = self.language.getRegister(self._registers[name])
        assert reg is not None, name
        return reg

    @classmethod
    def for_platform(cls, platform: platforms.Platform):
        """Find the appropriate MachineDef for your architecture

        Arguments:
            arch: The architecture ID you want
            byteorder: The byteorderness you want

        Returns:
            An instance of the appropriate MachineDef

        Raises:
            ValueError: If no MachineDef subclass matches your request
        """
        try:
            return utils.find_subclass(
                cls,
                lambda x: x.arch == platform.architecture
                and x.byteorder == platform.byteorder,
            )
        except:
            raise ValueError(f"No machine model for {platform}")



from ... import platforms, state
from . import cpu


class AArch64(cpu.CPU):
    """Auto-generated CPU state for aarch64:v8a:little

    Generated from Pcode language AARCH64:LE:64:v8A, and Unicorn package
    unicorn.arm64_const.
    """

    platform = platforms.Platform(
        platforms.Architecture.AARCH64, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__()
        # *** General Purpose Registers ***
        self.x0 = state.Register("x0", 8)
        self.add(self.x0)
        self.w0 = state.RegisterAlias("w0", self.x0, 4, 0)
        self.add(self.w0)
        self.x1 = state.Register("x1", 8)
        self.add(self.x1)
        self.w1 = state.RegisterAlias("w1", self.x1, 4, 0)
        self.add(self.w1)
        self.x2 = state.Register("x2", 8)
        self.add(self.x2)
        self.w2 = state.RegisterAlias("w2", self.x2, 4, 0)
        self.add(self.w2)
        self.x3 = state.Register("x3", 8)
        self.add(self.x3)
        self.w3 = state.RegisterAlias("w3", self.x3, 4, 0)
        self.add(self.w3)
        self.x4 = state.Register("x4", 8)
        self.add(self.x4)
        self.w4 = state.RegisterAlias("w4", self.x4, 4, 0)
        self.add(self.w4)
        self.x5 = state.Register("x5", 8)
        self.add(self.x5)
        self.w5 = state.RegisterAlias("w5", self.x5, 4, 0)
        self.add(self.w5)
        self.x6 = state.Register("x6", 8)
        self.add(self.x6)
        self.w6 = state.RegisterAlias("w6", self.x6, 4, 0)
        self.add(self.w6)
        self.x7 = state.Register("x7", 8)
        self.add(self.x7)
        self.w7 = state.RegisterAlias("w7", self.x7, 4, 0)
        self.add(self.w7)
        self.x8 = state.Register("x8", 8)
        self.add(self.x8)
        self.w8 = state.RegisterAlias("w8", self.x8, 4, 0)
        self.add(self.w8)
        self.x9 = state.Register("x9", 8)
        self.add(self.x9)
        self.w9 = state.RegisterAlias("w9", self.x9, 4, 0)
        self.add(self.w9)
        self.x10 = state.Register("x10", 8)
        self.add(self.x10)
        self.w10 = state.RegisterAlias("w10", self.x10, 4, 0)
        self.add(self.w10)
        self.x11 = state.Register("x11", 8)
        self.add(self.x11)
        self.w11 = state.RegisterAlias("w11", self.x10, 4, 0)
        self.add(self.w11)
        self.x12 = state.Register("x12", 8)
        self.add(self.x12)
        self.w12 = state.RegisterAlias("w12", self.x10, 4, 0)
        self.add(self.w12)
        self.x13 = state.Register("x13", 8)
        self.add(self.x13)
        self.w13 = state.RegisterAlias("w13", self.x13, 4, 0)
        self.add(self.w13)
        self.x14 = state.Register("x14", 8)
        self.add(self.x14)
        self.w14 = state.RegisterAlias("w14", self.x14, 4, 0)
        self.add(self.w14)
        self.x15 = state.Register("x15", 8)
        self.add(self.x15)
        self.w15 = state.RegisterAlias("w15", self.x15, 4, 0)
        self.add(self.w15)
        self.x16 = state.Register("x16", 8)
        self.add(self.x16)
        self.w16 = state.RegisterAlias("w16", self.x16, 4, 0)
        self.add(self.w16)
        self.x17 = state.Register("x17", 8)
        self.add(self.x17)
        self.w17 = state.RegisterAlias("w17", self.x17, 4, 0)
        self.add(self.w17)
        self.x18 = state.Register("x18", 8)
        self.add(self.x18)
        self.w18 = state.RegisterAlias("w18", self.x18, 4, 0)
        self.add(self.w18)
        self.x19 = state.Register("x19", 8)
        self.add(self.x19)
        self.w19 = state.RegisterAlias("w19", self.x19, 4, 0)
        self.add(self.w19)
        self.x20 = state.Register("x20", 8)
        self.add(self.x20)
        self.w20 = state.RegisterAlias("w20", self.x20, 4, 0)
        self.add(self.w20)
        self.x21 = state.Register("x21", 8)
        self.add(self.x21)
        self.w21 = state.RegisterAlias("w21", self.x20, 4, 0)
        self.add(self.w21)
        self.x22 = state.Register("x22", 8)
        self.add(self.x22)
        self.w22 = state.RegisterAlias("w22", self.x20, 4, 0)
        self.add(self.w22)
        self.x23 = state.Register("x23", 8)
        self.add(self.x23)
        self.w23 = state.RegisterAlias("w23", self.x23, 4, 0)
        self.add(self.w23)
        self.x24 = state.Register("x24", 8)
        self.add(self.x24)
        self.w24 = state.RegisterAlias("w24", self.x24, 4, 0)
        self.add(self.w24)
        self.x25 = state.Register("x25", 8)
        self.add(self.x25)
        self.w25 = state.RegisterAlias("w25", self.x25, 4, 0)
        self.add(self.w25)
        self.x26 = state.Register("x26", 8)
        self.add(self.x26)
        self.w26 = state.RegisterAlias("w26", self.x26, 4, 0)
        self.add(self.w26)
        self.x27 = state.Register("x27", 8)
        self.add(self.x27)
        self.w27 = state.RegisterAlias("w27", self.x27, 4, 0)
        self.add(self.w27)
        self.x28 = state.Register("x28", 8)
        self.add(self.x28)
        self.w28 = state.RegisterAlias("w28", self.x28, 4, 0)
        self.add(self.w28)
        self.x29 = state.Register("x29", 8)
        self.add(self.x29)
        self.w29 = state.RegisterAlias("w29", self.x29, 4, 0)
        self.add(self.w29)
        self.x30 = state.Register("x30", 8)
        self.add(self.x30)
        self.w30 = state.RegisterAlias("w30", self.x30, 4, 0)
        self.add(self.w30)
        # *** Program Counter ***
        self.pc = state.Register("pc", 8)
        self.add(self.pc)
        # *** Stack Pointer ***
        self.sp = state.Register("sp", 8)
        self.add(self.sp)
        self.wsp = state.RegisterAlias("wsp", self.sp, 4, 0)
        self.add(self.wsp)
        # *** Frame Pointer ***
        self.fp = state.RegisterAlias("fp", self.x29, 8, 0)
        self.add(self.fp)
        # *** Link Register ***
        self.lr = state.RegisterAlias("lr", self.x30, 8, 0)
        self.add(self.lr)
        # *** Zero Register ***
        self.xzr = state.FixedRegister("xzr", 8, 0)
        self.add(self.xzr)
        self.wzr = state.RegisterAlias("wzr", self.xzr, 4, 0)
        self.add(self.wzr)
        # *** System Control Registers ***
        # NOTE: "_elX" indicates that only exception level X or greater can access this register.
        # NOTE: This list is far from complete; it only covers what Unicorn supports
        # Condition Code Register
        self.fpcr = state.Register("fpcr", 8)
        self.add(self.fpcr)
        # Floating Point Status Register
        self.fpsr = state.Register("fpsr", 8)
        self.add(self.fpsr)
        # Banked stack pointers for exception handlers
        self.sp_el0 = state.Register("sp_el0", 8)
        self.add(self.sp_el0)
        self.sp_el1 = state.Register("sp_el1", 8)
        self.add(self.sp_el1)
        self.sp_el2 = state.Register("sp_el2", 8)
        self.add(self.sp_el2)
        self.sp_el3 = state.Register("sp_el3", 8)
        self.add(self.sp_el3)
        # Banked link registers for exception handlers
        # NOTE: Unicorn thinks there's an elr_el0; according to docs, it doesn't exist
        self.elr_el1 = state.Register("elr_el1", 8)
        self.add(self.elr_el1)
        self.elr_el2 = state.Register("elr_el2", 8)
        self.add(self.elr_el2)
        self.elr_el3 = state.Register("elr_el3", 8)
        self.add(self.elr_el3)
        # Banked exception syndrome registers for exception handlers
        # NOTE: Unicorn thinks there's a far_el0; according to docs, it doesn't exist
        self.far_el1 = state.Register("far_el1", 8)
        self.add(self.far_el1)
        self.far_el2 = state.Register("far_el2", 8)
        self.add(self.far_el2)
        self.far_el3 = state.Register("far_el3", 8)
        self.add(self.far_el3)
        # Banked vector base address registers for exception handlers
        # NOTE: vbar_el0 and vbar_el1 are aliases for each other.
        # Since vbar_el0 doesn't exist in angr, vbar_el1 has to be the "real" copy.
        self.vbar_el1 = state.Register("vbar_el1", 8)
        self.add(self.vbar_el1)
        self.vbar_el0 = state.RegisterAlias("vbar_el0", self.vbar_el1, 8, 0)
        self.add(self.vbar_el0)
        self.vbar_el2 = state.Register("vbar_el2", 8)
        self.add(self.vbar_el2)
        self.vbar_el3 = state.Register("vbar_el3", 8)
        self.add(self.vbar_el3)
        # Coprocessor access control register
        self.cpacr_el1 = state.Register("cpacr_el1", 8)
        self.add(self.cpacr_el1)
        # Memory Attribute Indirection Register
        self.mair_el1 = state.Register("mair_el1", 8)
        self.add(self.mair_el1)
        # Physical Address Register
        self.par_el1 = state.Register("par_el1", 8)
        self.add(self.par_el1)
        # Translation Table Zero Base Register
        self.ttbr0_el1 = state.Register("ttbr0_el1", 8)
        self.add(self.ttbr0_el1)
        # Translation Table One Base Register
        self.ttbr1_el1 = state.Register("ttbr1_el1", 8)
        self.add(self.ttbr1_el1)
        # Thread ID Register
        # NOTE: According to docs, there should be an el2 and el3 copy, too.
        self.tpidr_el0 = state.Register("tpidr_el0", 8)
        self.add(self.tpidr_el0)
        self.tpidr_el1 = state.Register("tpidr_el1", 8)
        self.add(self.tpidr_el1)
        # Userspace-visible Thread ID register
        self.tpidrro_el0 = state.Register("tpidrro_el0", 8)
        self.add(self.tpidrro_el0)
        # *** Floating Point Registers ***
        # Scalar Floating Point Registers
        self.q0 = state.Register("q0", 16)
        self.add(self.q0)
        self.d0 = state.RegisterAlias("d0", self.q0, 8, 0)
        self.add(self.d0)
        self.s0 = state.RegisterAlias("s0", self.q0, 4, 0)
        self.add(self.s0)
        self.h0 = state.RegisterAlias("h0", self.q0, 2, 0)
        self.add(self.h0)
        self.b0 = state.RegisterAlias("b0", self.q0, 1, 0)
        self.add(self.b0)
        self.q1 = state.Register("q1", 16)
        self.add(self.q1)
        self.d1 = state.RegisterAlias("d1", self.q1, 8, 0)
        self.add(self.d1)
        self.s1 = state.RegisterAlias("s1", self.q1, 4, 0)
        self.add(self.s1)
        self.h1 = state.RegisterAlias("h1", self.q1, 2, 0)
        self.add(self.h1)
        self.b1 = state.RegisterAlias("b1", self.q1, 1, 0)
        self.add(self.b1)
        self.q2 = state.Register("q2", 16)
        self.add(self.q2)
        self.d2 = state.RegisterAlias("d2", self.q2, 8, 0)
        self.add(self.d2)
        self.s2 = state.RegisterAlias("s2", self.q2, 4, 0)
        self.add(self.s2)
        self.h2 = state.RegisterAlias("h2", self.q2, 2, 0)
        self.add(self.h2)
        self.b2 = state.RegisterAlias("b2", self.q2, 1, 0)
        self.add(self.b2)
        self.q3 = state.Register("q3", 16)
        self.add(self.q3)
        self.d3 = state.RegisterAlias("d3", self.q3, 8, 0)
        self.add(self.d3)
        self.s3 = state.RegisterAlias("s3", self.q3, 4, 0)
        self.add(self.s3)
        self.h3 = state.RegisterAlias("h3", self.q3, 2, 0)
        self.add(self.h3)
        self.b3 = state.RegisterAlias("b3", self.q3, 1, 0)
        self.add(self.b3)
        self.q4 = state.Register("q4", 16)
        self.add(self.q4)
        self.d4 = state.RegisterAlias("d4", self.q4, 8, 0)
        self.add(self.d4)
        self.s4 = state.RegisterAlias("s4", self.q4, 4, 0)
        self.add(self.s4)
        self.h4 = state.RegisterAlias("h4", self.q4, 2, 0)
        self.add(self.h4)
        self.b4 = state.RegisterAlias("b4", self.q4, 1, 0)
        self.add(self.b4)
        self.q5 = state.Register("q5", 16)
        self.add(self.q5)
        self.d5 = state.RegisterAlias("d5", self.q5, 8, 0)
        self.add(self.d5)
        self.s5 = state.RegisterAlias("s5", self.q5, 4, 0)
        self.add(self.s5)
        self.h5 = state.RegisterAlias("h5", self.q5, 2, 0)
        self.add(self.h5)
        self.b5 = state.RegisterAlias("b5", self.q5, 1, 0)
        self.add(self.b5)
        self.q6 = state.Register("q6", 16)
        self.add(self.q6)
        self.d6 = state.RegisterAlias("d6", self.q6, 8, 0)
        self.add(self.d6)
        self.s6 = state.RegisterAlias("s6", self.q6, 4, 0)
        self.add(self.s6)
        self.h6 = state.RegisterAlias("h6", self.q6, 2, 0)
        self.add(self.h6)
        self.b6 = state.RegisterAlias("b6", self.q6, 1, 0)
        self.add(self.b6)
        self.q7 = state.Register("q7", 16)
        self.add(self.q7)
        self.d7 = state.RegisterAlias("d7", self.q7, 8, 0)
        self.add(self.d7)
        self.s7 = state.RegisterAlias("s7", self.q7, 4, 0)
        self.add(self.s7)
        self.h7 = state.RegisterAlias("h7", self.q7, 2, 0)
        self.add(self.h7)
        self.b7 = state.RegisterAlias("b7", self.q7, 1, 0)
        self.add(self.b7)
        self.q8 = state.Register("q8", 16)
        self.add(self.q8)
        self.d8 = state.RegisterAlias("d8", self.q8, 8, 0)
        self.add(self.d8)
        self.s8 = state.RegisterAlias("s8", self.q8, 4, 0)
        self.add(self.s8)
        self.h8 = state.RegisterAlias("h8", self.q8, 2, 0)
        self.add(self.h8)
        self.b8 = state.RegisterAlias("b8", self.q8, 1, 0)
        self.add(self.b8)
        self.q9 = state.Register("q9", 16)
        self.add(self.q9)
        self.d9 = state.RegisterAlias("d9", self.q9, 8, 0)
        self.add(self.d9)
        self.s9 = state.RegisterAlias("s9", self.q9, 4, 0)
        self.add(self.s9)
        self.h9 = state.RegisterAlias("h9", self.q9, 2, 0)
        self.add(self.h9)
        self.b9 = state.RegisterAlias("b9", self.q9, 1, 0)
        self.add(self.b9)
        self.q10 = state.Register("q10", 16)
        self.add(self.q10)
        self.d10 = state.RegisterAlias("d10", self.q10, 8, 0)
        self.add(self.d10)
        self.s10 = state.RegisterAlias("s10", self.q10, 4, 0)
        self.add(self.s10)
        self.h10 = state.RegisterAlias("h10", self.q10, 2, 0)
        self.add(self.h10)
        self.b10 = state.RegisterAlias("b10", self.q10, 1, 0)
        self.add(self.b10)
        self.q11 = state.Register("q11", 16)
        self.add(self.q11)
        self.d11 = state.RegisterAlias("d11", self.q11, 8, 0)
        self.add(self.d11)
        self.s11 = state.RegisterAlias("s11", self.q11, 4, 0)
        self.add(self.s11)
        self.h11 = state.RegisterAlias("h11", self.q11, 2, 0)
        self.add(self.h11)
        self.b11 = state.RegisterAlias("b11", self.q11, 1, 0)
        self.add(self.b11)
        self.q12 = state.Register("q12", 16)
        self.add(self.q12)
        self.d12 = state.RegisterAlias("d12", self.q12, 8, 0)
        self.add(self.d12)
        self.s12 = state.RegisterAlias("s12", self.q12, 4, 0)
        self.add(self.s12)
        self.h12 = state.RegisterAlias("h12", self.q12, 2, 0)
        self.add(self.h12)
        self.b12 = state.RegisterAlias("b12", self.q12, 1, 0)
        self.add(self.b12)
        self.q13 = state.Register("q13", 16)
        self.add(self.q13)
        self.d13 = state.RegisterAlias("d13", self.q13, 8, 0)
        self.add(self.d13)
        self.s13 = state.RegisterAlias("s13", self.q13, 4, 0)
        self.add(self.s13)
        self.h13 = state.RegisterAlias("h13", self.q13, 2, 0)
        self.add(self.h13)
        self.b13 = state.RegisterAlias("b13", self.q13, 1, 0)
        self.add(self.b13)
        self.q14 = state.Register("q14", 16)
        self.add(self.q14)
        self.d14 = state.RegisterAlias("d14", self.q14, 8, 0)
        self.add(self.d14)
        self.s14 = state.RegisterAlias("s14", self.q14, 4, 0)
        self.add(self.s14)
        self.h14 = state.RegisterAlias("h14", self.q14, 2, 0)
        self.add(self.h14)
        self.b14 = state.RegisterAlias("b14", self.q14, 1, 0)
        self.add(self.b14)
        self.q15 = state.Register("q15", 16)
        self.add(self.q15)
        self.d15 = state.RegisterAlias("d15", self.q15, 8, 0)
        self.add(self.d15)
        self.s15 = state.RegisterAlias("s15", self.q15, 4, 0)
        self.add(self.s15)
        self.h15 = state.RegisterAlias("h15", self.q15, 2, 0)
        self.add(self.h15)
        self.b15 = state.RegisterAlias("b15", self.q15, 1, 0)
        self.add(self.b15)
        self.q16 = state.Register("q16", 16)
        self.add(self.q16)
        self.d16 = state.RegisterAlias("d16", self.q16, 8, 0)
        self.add(self.d16)
        self.s16 = state.RegisterAlias("s16", self.q16, 4, 0)
        self.add(self.s16)
        self.h16 = state.RegisterAlias("h16", self.q16, 2, 0)
        self.add(self.h16)
        self.b16 = state.RegisterAlias("b16", self.q16, 1, 0)
        self.add(self.b16)
        self.q17 = state.Register("q17", 16)
        self.add(self.q17)
        self.d17 = state.RegisterAlias("d17", self.q17, 8, 0)
        self.add(self.d17)
        self.s17 = state.RegisterAlias("s17", self.q17, 4, 0)
        self.add(self.s17)
        self.h17 = state.RegisterAlias("h17", self.q17, 2, 0)
        self.add(self.h17)
        self.b17 = state.RegisterAlias("b17", self.q17, 1, 0)
        self.add(self.b17)
        self.q18 = state.Register("q18", 16)
        self.add(self.q18)
        self.d18 = state.RegisterAlias("d18", self.q18, 8, 0)
        self.add(self.d18)
        self.s18 = state.RegisterAlias("s18", self.q18, 4, 0)
        self.add(self.s18)
        self.h18 = state.RegisterAlias("h18", self.q18, 2, 0)
        self.add(self.h18)
        self.b18 = state.RegisterAlias("b18", self.q18, 1, 0)
        self.add(self.b18)
        self.q19 = state.Register("q19", 16)
        self.add(self.q19)
        self.d19 = state.RegisterAlias("d19", self.q19, 8, 0)
        self.add(self.d19)
        self.s19 = state.RegisterAlias("s19", self.q19, 4, 0)
        self.add(self.s19)
        self.h19 = state.RegisterAlias("h19", self.q19, 2, 0)
        self.add(self.h19)
        self.b19 = state.RegisterAlias("b19", self.q19, 1, 0)
        self.add(self.b19)
        self.q20 = state.Register("q20", 16)
        self.add(self.q20)
        self.d20 = state.RegisterAlias("d20", self.q20, 8, 0)
        self.add(self.d20)
        self.s20 = state.RegisterAlias("s20", self.q20, 4, 0)
        self.add(self.s20)
        self.h20 = state.RegisterAlias("h20", self.q20, 2, 0)
        self.add(self.h20)
        self.b20 = state.RegisterAlias("b20", self.q20, 1, 0)
        self.add(self.b20)
        self.q21 = state.Register("q21", 16)
        self.add(self.q21)
        self.d21 = state.RegisterAlias("d21", self.q21, 8, 0)
        self.add(self.d21)
        self.s21 = state.RegisterAlias("s21", self.q21, 4, 0)
        self.add(self.s21)
        self.h21 = state.RegisterAlias("h21", self.q21, 2, 0)
        self.add(self.h21)
        self.b21 = state.RegisterAlias("b21", self.q21, 1, 0)
        self.add(self.b21)
        self.q22 = state.Register("q22", 16)
        self.add(self.q22)
        self.d22 = state.RegisterAlias("d22", self.q22, 8, 0)
        self.add(self.d22)
        self.s22 = state.RegisterAlias("s22", self.q22, 4, 0)
        self.add(self.s22)
        self.h22 = state.RegisterAlias("h22", self.q22, 2, 0)
        self.add(self.h22)
        self.b22 = state.RegisterAlias("b22", self.q22, 1, 0)
        self.add(self.b22)
        self.q23 = state.Register("q23", 16)
        self.add(self.q23)
        self.d23 = state.RegisterAlias("d23", self.q23, 8, 0)
        self.add(self.d23)
        self.s23 = state.RegisterAlias("s23", self.q23, 4, 0)
        self.add(self.s23)
        self.h23 = state.RegisterAlias("h23", self.q23, 2, 0)
        self.add(self.h23)
        self.b23 = state.RegisterAlias("b23", self.q23, 1, 0)
        self.add(self.b23)
        self.q24 = state.Register("q24", 16)
        self.add(self.q24)
        self.d24 = state.RegisterAlias("d24", self.q24, 8, 0)
        self.add(self.d24)
        self.s24 = state.RegisterAlias("s24", self.q24, 4, 0)
        self.add(self.s24)
        self.h24 = state.RegisterAlias("h24", self.q24, 2, 0)
        self.add(self.h24)
        self.b24 = state.RegisterAlias("b24", self.q24, 1, 0)
        self.add(self.b24)
        self.q25 = state.Register("q25", 16)
        self.add(self.q25)
        self.d25 = state.RegisterAlias("d25", self.q25, 8, 0)
        self.add(self.d25)
        self.s25 = state.RegisterAlias("s25", self.q25, 4, 0)
        self.add(self.s25)
        self.h25 = state.RegisterAlias("h25", self.q25, 2, 0)
        self.add(self.h25)
        self.b25 = state.RegisterAlias("b25", self.q25, 1, 0)
        self.add(self.b25)
        self.q26 = state.Register("q26", 16)
        self.add(self.q26)
        self.d26 = state.RegisterAlias("d26", self.q26, 8, 0)
        self.add(self.d26)
        self.s26 = state.RegisterAlias("s26", self.q26, 4, 0)
        self.add(self.s26)
        self.h26 = state.RegisterAlias("h26", self.q26, 2, 0)
        self.add(self.h26)
        self.b26 = state.RegisterAlias("b26", self.q26, 1, 0)
        self.add(self.b26)
        self.q27 = state.Register("q27", 16)
        self.add(self.q27)
        self.d27 = state.RegisterAlias("d27", self.q27, 8, 0)
        self.add(self.d27)
        self.s27 = state.RegisterAlias("s27", self.q27, 4, 0)
        self.add(self.s27)
        self.h27 = state.RegisterAlias("h27", self.q27, 2, 0)
        self.add(self.h27)
        self.b27 = state.RegisterAlias("b27", self.q27, 1, 0)
        self.add(self.b27)
        self.q28 = state.Register("q28", 16)
        self.add(self.q28)
        self.d28 = state.RegisterAlias("d28", self.q28, 8, 0)
        self.add(self.d28)
        self.s28 = state.RegisterAlias("s28", self.q28, 4, 0)
        self.add(self.s28)
        self.h28 = state.RegisterAlias("h28", self.q28, 2, 0)
        self.add(self.h28)
        self.b28 = state.RegisterAlias("b28", self.q28, 1, 0)
        self.add(self.b28)
        self.q29 = state.Register("q29", 16)
        self.add(self.q29)
        self.d29 = state.RegisterAlias("d29", self.q29, 8, 0)
        self.add(self.d29)
        self.s29 = state.RegisterAlias("s29", self.q29, 4, 0)
        self.add(self.s29)
        self.h29 = state.RegisterAlias("h29", self.q29, 2, 0)
        self.add(self.h29)
        self.b29 = state.RegisterAlias("b29", self.q29, 1, 0)
        self.add(self.b29)
        self.q30 = state.Register("q30", 16)
        self.add(self.q30)
        self.d30 = state.RegisterAlias("d30", self.q30, 8, 0)
        self.add(self.d30)
        self.s30 = state.RegisterAlias("s30", self.q30, 4, 0)
        self.add(self.s30)
        self.h30 = state.RegisterAlias("h30", self.q30, 2, 0)
        self.add(self.h30)
        self.b30 = state.RegisterAlias("b30", self.q30, 1, 0)
        self.add(self.b30)
        self.q31 = state.Register("q31", 16)
        self.add(self.q31)
        self.d31 = state.RegisterAlias("d31", self.q31, 8, 0)
        self.add(self.d31)
        self.s31 = state.RegisterAlias("s31", self.q31, 4, 0)
        self.add(self.s31)
        self.h31 = state.RegisterAlias("h31", self.q31, 2, 0)
        self.add(self.h31)
        self.b31 = state.RegisterAlias("b31", self.q31, 1, 0)
        self.add(self.b31)
        # Vector registers
        # TODO: Figure out how to model these



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import RiscV64SysVModel


class RiscV64SysVAbs(Abs, RiscV64SysVModel):
    pass


class RiscV64SysVLAbs(LAbs, RiscV64SysVModel):
    pass


class RiscV64SysVLLAbs(LLAbs, RiscV64SysVModel):
    pass


class RiscV64SysVAtof(Atof, RiscV64SysVModel):
    pass


class RiscV64SysVAtoi(Atoi, RiscV64SysVModel):
    pass


class RiscV64SysVAtol(Atol, RiscV64SysVModel):
    pass


class RiscV64SysVAtoll(Atoll, RiscV64SysVModel):
    pass


class RiscV64SysVCalloc(Calloc, RiscV64SysVModel):
    pass


class RiscV64SysVDiv(Div, RiscV64SysVModel):
    pass


class RiscV64SysVLDiv(LDiv, RiscV64SysVModel):
    pass


class RiscV64SysVLLDiv(LLDiv, RiscV64SysVModel):
    pass


class RiscV64SysVExit(Exit, RiscV64SysVModel):
    pass


class RiscV64SysVFree(Free, RiscV64SysVModel):
    pass


class RiscV64SysVMalloc(Malloc, RiscV64SysVModel):
    pass


class RiscV64SysVQSort(QSort, RiscV64SysVModel):
    pass


class RiscV64SysVRand(Rand, RiscV64SysVModel):
    pass


class RiscV64SysVRealloc(Realloc, RiscV64SysVModel):
    pass


class RiscV64SysVSrand(Srand, RiscV64SysVModel):
    pass


class RiscV64SysVAbort(Abort, RiscV64SysVModel):
    pass


class RiscV64SysVAtexit(Atexit, RiscV64SysVModel):
    pass


class RiscV64SysVGetenv(Getenv, RiscV64SysVModel):
    pass


class RiscV64SysVMblen(Mblen, RiscV64SysVModel):
    pass


class RiscV64SysVMbstowcs(Mbstowcs, RiscV64SysVModel):
    pass


class RiscV64SysVMbtowc(Mbtowc, RiscV64SysVModel):
    pass


class RiscV64SysVSystem(System, RiscV64SysVModel):
    pass


class RiscV64SysVWcstombs(Wcstombs, RiscV64SysVModel):
    pass


class RiscV64SysVWctomb(Wctomb, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVAbs",
    "RiscV64SysVLAbs",
    "RiscV64SysVLLAbs",
    "RiscV64SysVAtof",
    "RiscV64SysVAtoi",
    "RiscV64SysVAtol",
    "RiscV64SysVAtoll",
    "RiscV64SysVCalloc",
    "RiscV64SysVDiv",
    "RiscV64SysVLDiv",
    "RiscV64SysVLLDiv",
    "RiscV64SysVExit",
    "RiscV64SysVFree",
    "RiscV64SysVMalloc",
    "RiscV64SysVQSort",
    "RiscV64SysVRand",
    "RiscV64SysVRealloc",
    "RiscV64SysVSrand",
    "RiscV64SysVAbort",
    "RiscV64SysVAtexit",
    "RiscV64SysVGetenv",
    "RiscV64SysVMblen",
    "RiscV64SysVMbstowcs",
    "RiscV64SysVMbtowc",
    "RiscV64SysVSystem",
    "RiscV64SysVWcstombs",
    "RiscV64SysVWctomb",
]



import dataclasses
import typing

from ... import hinting


class ClaripySerializable(hinting.Hint):
    """Serializable wrapper that allows serializing claripy expressions in hints."""

    @classmethod
    def claripy_to_dict(cls, expr):
        if (
            expr is None
            or isinstance(expr, int)
            or isinstance(expr, str)
            or isinstance(expr, bool)
        ):
            return expr
        else:
            return {"op": expr.op, "args": list(map(cls.claripy_to_dict, expr.args))}

    def __init__(self, expr):
        self.expr = expr

    def to_dict(self):
        return self.claripy_to_dict(self.expr)

    @classmethod
    def from_dict(cls, dict):
        raise NotImplementedError(
            "Rebuilding clairpy is a lot harder than tearing it apart"
        )


@dataclasses.dataclass(frozen=True)
class FieldHint(hinting.Hint):
    address: int = 0
    size: int = 0

    def pp(self, out):
        out(self.message)
        out(f"  address:    {hex(self.address)}")
        out(f"  size:       {self.size}")


@dataclasses.dataclass(frozen=True)
class TrackedFieldHint(FieldHint):
    label: str = ""

    def pp(self, out):
        super().pp(out)
        out(f"  label:      {self.label}")


@dataclasses.dataclass(frozen=True)
class FieldEventHint(FieldHint):
    """Hint representing a detected, unhandled field"""

    pc: int = 0
    guards: typing.List[typing.Tuple[int, ClaripySerializable]] = dataclasses.field(
        default_factory=list
    )
    access: str = ""

    def pp(self, out):
        super().pp(out)
        out(f"  pc:         {hex(self.pc)}")
        out(f"  access:     {self.access}")
        out("  guards:")
        for pc, guard in self.guards:
            out(
                f"    {hex(pc)}: {guard.expr.shallow_repr()} [{list(filter(lambda x: x.op != 'BVV', guard.expr.leaf_asts()))}]"
            )


@dataclasses.dataclass(frozen=True)
class UnknownFieldHint(FieldEventHint):
    """Hint representing an access of an unknown field.

    The expression loaded from memory is not
    a simple slice of any known label.
    """

    message: str = "Accessed unknown field"  # Don't need to replace
    expr: str = ""

    def pp(self, out):
        super().pp(out)
        out(f"  expr:       {self.expr}")


@dataclasses.dataclass(frozen=True)
class PartialByteFieldAccessHint(FieldEventHint):
    """Hint representing an access to part of a known field."""

    message: str = "Partial access to known field"  # Don't need to replace
    label: str = ""
    start: int = 0
    end: int = 0

    def pp(self, out):
        super().pp(out)
        out(f"  field:      {self.label}[{self.start}:{self.end}]")


@dataclasses.dataclass(frozen=True)
class PartialByteFieldWriteHint(PartialByteFieldAccessHint):
    """Hint representing a write to part of a known field.

    Also includes the data I'm trying to write
    """

    message: str = "Partial write to known field"  # Don't need to replace
    access: str = "write"  # Don't need to replace
    expr: str = ""

    def pp(self, out):
        super().pp(out)
        out(f"  expr:       {self.expr}")


@dataclasses.dataclass(frozen=True)
class PartialBitFieldAccessHint(FieldEventHint):
    """Hint representing an access to a bit field within a known field."""

    message: str = "Bit field access in known field"  # Don't need to replace
    label: str = ""
    start: int = 0
    end: int = 0

    def pp(self, out):
        super().pp(out)
        out(f"  field:      {self.label}[{self.start}:{self.end}]")



import io
import sys
import typing

from ...exceptions import ConfigurationError
from ...platforms import ABI, Platform


class FDIOError(Exception):
    """Exception indicating an error case in the file IO model"""

    pass


class FileDescriptor:
    """File Descriptor Representation"""

    def __init__(
        self,
        name: str,
        readable: bool = False,
        writable: bool = False,
        seekable: bool = False,
    ):
        self.name = name

        self.readable = readable
        self.writable = writable
        self.seekable = seekable

        self.ungetc_buf = b""

        self.cursor = 0

        self.eof = False

    @property
    def _backing(self) -> typing.IO:
        # Files aren't pickleable,
        # and anyway we don't want contention on one file object.
        # Instead, dynamically produce the backing file-like,
        # with the understanding that it may need special handling.
        raise FDIOError("File {self.name} has no backing")

    def read(self, size: int, ungetc: bool = False) -> bytes:
        """Read data from this file descriptor

        Arguments:
            size: Number of bytes to read
            ungetc: Set to true to use the 'ungetc' buffer
        """
        if not self.readable:
            raise FDIOError(f"File {self.name} is not readable")

        # NOTE: This integrates support for ungetc.
        # This feature is technically part of FILE * objects;
        # ungetc effectively appends bytes to the beg
        ungetc_data = b""
        if ungetc and len(self.ungetc_buf) > 0:
            if size == -1:
                ungetc_data = self.ungetc_buf
                self.ungetc_buf = b""
            else:
                ungetc_data = self.ungetc_buf[0:size]
                self.ungetc_buf = self.ungetc_buf[size:]

                size -= len(ungetc_data)

        file = self._backing

        if self.seekable:
            # File is seekable; mimic it against the back end.
            file.seek(self.cursor, 0)

        data: typing.Union[str, bytes] = b""
        if size != 0:
            data = file.read(size)
            self.cursor += len(data)
            if len(data) != size:
                print("EOF")
                self.eof = True

        if isinstance(data, str):
            # This is a text file.  Need to encode the result
            assert hasattr(file, "encoding")
            return ungetc_data + data.encode(file.encoding)
        else:
            return ungetc_data + data

    def read_string(self, size: int = -1) -> bytes:
        """Read a newline-terminated string from this file

        Arguments
            size: Maximum number of bytes to read.  Defaults to as many as possible

        Returns:
            String read from file
        """
        out = b""
        size -= 1
        while size != 0:
            c = self.read(1)
            if len(c) < 1:
                break
            out += c

            if c == b"\n":
                break

            if size > 0:
                size -= 1
        out += b"\0"
        return out

    def write(self, data: bytes) -> None:
        """Write data to this file

        Arguments:
            data: Bytes to write to this file
        """
        if not self.writable:
            raise FDIOError(f"File {self.name} is not writable")

        # TODO: Model the effects of ungetc.
        # By observation, mixing writes with ungetc is unsupported;
        # the test program segfaulted when I wran fputc after ungetc

        file = self._backing

        if self.seekable:
            file.seek(self.cursor, 0)

        if hasattr(file, "encoding"):
            # File is a text file.  Need to encode the results
            file.write(data.decode(file.encoding))
        else:
            file.write(data)

        self.cursor += len(data)

    def seek(self, pos: int, whence: int) -> int:
        """Set the cursor for this file

        Arguments:
            pos: Position, possibly relative
            whence: How to interpret 'pos'.  See real implementations for possible values
        """
        if not self.seekable:
            raise FDIOError(f"File {self.name} is not seekable")

        if whence == 0:
            self.cursor = pos
        elif whence == 1:
            self.cursor += pos
        elif whence == 2:
            # Figure out how to find the end of the stream
            raise NotImplementedError()
        else:
            raise FDIOError(f"Unknown 'whence' {whence} when seeking {self.name}")

        return self.cursor

    def ungetc(self, char: int) -> None:
        """Push a character back to the buffer

        In real life, this is only supported by stdio FILE* objects.
        It pushes the character to an internal buffer that's
        read from before the actual stream gets read.

        Its interactions with 'write', 'seek', and readingf from 'cursor'
        are all undefined.

        Arguments:
            char: Character to push
        """
        if not self.readable:
            raise FDIOError(f"File {self.name} is not readable")

        self.ungetc_buf = bytes([char]) + self.ungetc_buf


class BytesFileDescriptor(FileDescriptor):
    """File descriptor backed by a byte string."""

    def __init__(
        self,
        name: str,
        initial: typing.Union[bytes, io.BytesIO] = b"",
        readable: bool = False,
        writable: bool = False,
        seekable: bool = False,
    ):
        super().__init__(name, readable=readable, writable=writable, seekable=seekable)
        if isinstance(initial, bytes):
            initial = io.BytesIO(initial)
        self.backing_io = initial

    @property
    def _backing(self) -> typing.IO:
        return self.backing_io


class StdinFileDescriptor(FileDescriptor):
    """File descriptor backed by host's stdin.

    Read-only, not seekable.
    """

    def __init__(self):
        super().__init__("stdin", readable=True)

    @property
    def _backing(self) -> typing.IO:
        return sys.stdin


class StdoutFileDescriptor(FileDescriptor):
    """File descriptor backed by host's stdout.

    Write-only, not seekable.
    """

    def __init__(self):
        super().__init__("stdout", writable=True)

    @property
    def _backing(self) -> typing.IO:
        return sys.stdout


class StderrFileDescriptor(FileDescriptor):
    """File descriptor backed by host's stderr.

    Write-only, not seekable.
    """

    def __init__(self):
        super().__init__("stderr", writable=True)

    @property
    def _backing(self) -> typing.IO:
        return sys.stderr


class FileDescriptorManager:
    """Manager for wrangling file descriptors

    This handles open, close, and other file descriptor
    manipulation operations.  It also lets you access the file descriptor
    representations themselves.

    This is a unified model that supports C99,
    and is pretty much identical between System V and Windows.

    It also supports an extermely basic filesystem model.
    It's disabled by default, and enabled by setting 'model_fs' to True.
    Its utility is currently extremely limited;
    it has no directory tree or permissions model,
    but it can help for very basic interactions.

    There can absolutely be ABI-specific subclasses.
    """

    _singletons: typing.Dict[
        typing.Tuple[Platform, ABI], "FileDescriptorManager"
    ] = dict()

    def __init__(self):
        # Enables full FS modeling.
        # By default, this will
        self.model_fs = False

        self._fds = dict()
        self._files = dict()

        # Allocate stdstreams
        # NOTE: This is POSIX-specific
        self._fds[0] = StdinFileDescriptor()
        self._fds[1] = StdoutFileDescriptor()
        self._fds[2] = StderrFileDescriptor()

    def add_file(self, name: str, init: bytes = b"") -> None:
        """Add a file to the manager

        This allows basic support for opening files
        without creating them.

        As of now, there is no actual file tree support;
        whatever string is used as the name is what will match.

        Arguments:
            name: Name or path of the file to place
            init: Initial bytes contained in that file
        """
        if not self.model_fs:
            raise ConfigurationError("Full FS support not enabled.")

        self._files[name] = io.BytesIO(init)

    def get_file(self, name: str) -> io.BytesIO:
        """Get the backing stream behind a file

        Arguments:
            name: The name to look up

        Returns:
            The file-like object storing the byte stream.
        """
        if name in self._files:
            return self._files[name]

        raise KeyError(f"No file named {name}")

    def open(
        self,
        name: str,
        readable: bool,
        writable: bool,
        create: bool,
        truncate: bool,
        append: bool,
        seekable: bool = True,
    ) -> int:
        """Mimic opening a file

        This will behave a bit differently depending
        on whether 'self.model_fs' is set.

        If it's unset, it will create file descriptors blindly,
        and attempting to interact with them will raise exceptions.

        If it's set, it will use the basic file model; see add_file() and get_file().

        As of now, there is no actual file tree support;
        the name is merely used as a label.

        Arguments:
            name: Name of the file to open
            readable: Whether the file should be readable
            writable: Whether the file should be writable
            create: Whether the file should be created if it doesn't exist.
            truncate: Whether the file should be truncated when opened
            append: Whether the cursor should be set at the end of the file
            seekable: Whether the file should be seekable

        Returns:
            An integer file descriptor
        """

        # Limited to 256 file descriptors thanks to shenanigans with FILE * management.
        # If you need more than 256 file descriptors in a micro-execution context,
        # I have many questions.
        for fd in range(0, 1 << 8):
            if fd not in self._fds:
                if self.model_fs:
                    if name not in self._files:
                        if not create:
                            raise FDIOError(
                                f"{name} does not exist, and creation not specified"
                            )
                        self.add_file(name, b"")

                    initial = self._files[name]

                    if truncate:
                        initial.truncate()

                    self._fds[fd] = BytesFileDescriptor(
                        name,
                        initial,
                        readable=readable,
                        writable=writable,
                        seekable=seekable,
                    )

                    if append:
                        self._fds[fd].cursor = len(initial.getvalue())
                else:
                    self._fds[fd] = FileDescriptor(
                        name, readable=readable, writable=writable, seekable=seekable
                    )
                return fd
        raise FDIOError("Ran out of fds")

    def close(self, fd: int) -> None:
        """Close a file

        Arguments:
            fd: Integer file descriptor
        """
        if fd not in self._fds:
            raise FDIOError(f"Unknown fd {fd}")

        del self._fds[fd]

    def get(self, fd: int) -> FileDescriptor:
        """Get the file representaiton by its integer file descriptor

        Arguments:
            fd: The integer file descriptor

        Returns:
            The file representation tied to 'fd'.
        """
        if fd not in self._fds:
            raise FDIOError(f"Unknown fd {fd}")

        return self._fds[fd]

    def rename(self, old: str, new: str) -> None:
        """Rename a file.

        Only available if model_fs is set.

        As of now, there is no actual file tree support;
        the name is merely used as a label.

        Arguments:
            old: The current name of the file
            new: The new name of the file
        """
        if not self.model_fs:
            return

        if old not in self._files:
            raise FDIOError(f"Unknown file {old}")
        self._files[new] = self._files[old]
        del self._files[old]

    def remove(self, name: str) -> bool:
        """Remove a file

        Is a no-op if model_fs is not set.

        As of now, there is no actual file tree support;
        the name is merely used as a label.

        Arguments:
            name: Name of the file to remove
        Returns:
            True if model_fs is not set, or if the file was removed
        """

        if self.model_fs:
            if name in self._files:
                del self._files[name]
                return True
            return False
        else:
            return True

    # Magic number for identifying FILE * pointers created by this API.
    # ASCII for 'FI*'
    #
    # FILE * created by these models don't point to real structs.
    # Rather, they're an encoded form of the file descriptor;
    # the fd number is the LSB, with magic in bytes 1 through 3.
    #
    # On a 64-bit ABI, the upper four bytes should be zero;
    # I don't want to write two models for this.
    #
    # Functions that allow for multiple FILE * structs per FD
    # can go jump in a puddle.
    filestar_magic = 0x47492A

    @classmethod
    def filestar_to_fd(cls, ptr: int) -> int:
        """Convert a FILE* pointer to an integer file descriptor

        This will attempt to detect if the FILE* was actually created
        by SmallWorld's model.  If not, it raises an exception.

        Arguments:
            ptr: The FILE* pointer to decode

        Returns:
            The associated integer file descriptor
        """
        if (ptr >> 8) != cls.filestar_magic:
            raise FDIOError(f"FILE * {hex(ptr)} is not a FILE * created by this model.")

        return ptr & 0xFF

    @classmethod
    def fd_to_filestar(cls, fd: int) -> int:
        """Convert an integer file descriptor to a FILE* pointer

        Arguments:
            fd: The integer file descriptor to encode

        Returns:
            The associated FILE* pointer
        """
        return cls.filestar_magic << 8 | fd

    @classmethod
    def for_platform(cls, platform: Platform, abi: ABI):
        """Get an instance of this class for the desired platform

        NOTE: This isn't a true singleton, and I want it that way.
        Everything that asks for a manager during setup
        should get the same instance,
        but deep-copies of Machines should get their own managers

        Arguments:
            platform: The desired platform
            abi: The desired ABI

        Returns:
            An instance of the manager for the platform
        """
        if (platform, abi) not in cls._singletons:
            # TODO: Actually implement this when I have multiple implementations
            cls._singletons[(platform, abi)] = cls()
        return cls._singletons[(platform, abi)]



from .aarch64 import AArch64MachineDef
from .amd64 import AMD64MachineDef
from .arm import (  # ARMv6MMachineDef,; ARMv6MThumbMachineDef,
    ARMv5TMachineDef,
    ARMv7MMachineDef,
)
from .i386 import i386MachineDef
from .machdef import PandaMachineDef
from .mips import MIPSBEMachineDef, MIPSELMachineDef
from .mips64 import MIPS64BEMachineDef, MIPS64ELMachineDef
from .ppc import PowerPC32MachineDef  # , PowerPC64MachineDef

__all__ = [
    "AArch64MachineDef",
    "AMD64MachineDef",
    "PandaMachineDef",
    "ARMv5TMachineDef",
    #    "ARMv6MMachineDef",
    #    "ARMv6MThumbMachineDef",
    "ARMv7MMachineDef",
    "i386MachineDef",
    "MIPSBEMachineDef",
    "MIPSELMachineDef",
    "MIPS64BEMachineDef",
    "MIPS64ELMachineDef",
    "PowerPC32MachineDef",
    # "PowerPC64MachineDef",
]
